#THE CLASSES USED BY THE NICHEWEBTESTER AND THE REALWEBTESTER TO GENERATE GENERALIZED MODEL FOOD WEBS
#NORMALLY, A USER DOES NOT NEED TO CHANGE THE EXISTING CODE. IF YOU WANT TO ADD YOUR OWN GM PARAMETERIZATION, SIMPLY ADD IT TO THOSE PRESENT AT THE VERY END OF THIS FILE
####################################################################
#(C) BY H.AUFDERHEIDE 
# THIS SCRIPT IS DISTRIBUTED AS-IS
# YOU USE IT AT YOUR OWN RISK
#
## This work is licensed under the Creative Commons Attribution-NonCommercial 3.0 Unported License.
# To view a copy of this license, visit http://creativecommons.org/licenses/by-nc/3.0/ or
# send a letter to Creative Commons, 444 Castro Street, Suite 900, Mountain View, California, 94041, USA.
#
# The author takes no responsability whatshowever for the correctness and correct function of this script
# and is not liable for any damage caused through its correct or incorrect use.
######################################################################

#The class FoodWebAdjacency contains or generates a Food Web Adjacency Matrix for use in a generalized model.
# If makeNicheFW(<species>,<connectance>) is called it generates a Niche Model Food Web Adacency Matrix
# For this it uses an implementation of the "Niche Model" presented in "Simple rules yield complex food webs", Science 404,by R.J. Williams and N.D. Martinez, 2000
# If parseExternalAdj(<filename>) is called, a real-world adjacency in <filename> can be loaded.

#The Class FoodWebModel parametrizes a topology in self.topo in terms of the Generalized Modelling Parameters
# presented in "Generalized models as a universal approach to the analysis of nonlinear dynamical systems", Phys. Rev. E. 73, by T. Gross and U. Feudel, 2006
# Currently the Parametrization is either Randomly by calling makeRandomModel() or with constant dummy values by calling makeDummyModel().
# See the functions for the ranges and values used.

#The approach is detailed in "Generalized Models Reveal Stabilizing Factors in Food Webs", Science 325, by T. Gross, L. Rudolf, S. Levin, U. Dieckmann, 2009.


############################################################################################################################################################

import numpy as np
from scipy.linalg import eigvals,solve

############################################################################################################################################################
#A CLASS TO READ A TOPOLOGY (OR TO GENERATE NICHE MODEL TOPOLOGIES with connectance C and N species) (currently N>15 becomes pretty slow, due to pythons limitations)

class FoodWebAdjacency:	
##HELPER FUNCTIONS FOR ALL TOPOLOGIES
    def __init__(self):		#A COMPLETELY UNSPECIFIED AND EMPTY TOPOLOGY WITH THE FOLLOWING VARIABLES
      self.N=None		# Number of Species
      self.P=0			# Perturbation if required
      self.C=None		# Connectance (Only Niche Model)
      
      #Niche model parameters:
      self.c=None		# Predation Range Center (only Niche Model)
      self.n=None		# Niche Value (only Niche Model)
      self.r=None		# Predation Range Size (only Niche Model)

      #Always available output parameters
      self.Adjacency=None	# The Adjacency Matrix for use by the GM Model
      self.primprod=None	# Array indicating if a species is primary producer (primprod[i]=1 and 0 else)
      self.toppred=None		# Array indicating if a species is a top predator (toppred[i]=1 and 0 else)



    def _TestConnect(self):  #HELPER: A very basic connectance tester for Link Matrices
      connlist=np.zeros(len(self.Adjacency),dtype=np.int8)
      def _Mark(a):
	connlist[a]=1
	for newa in range(len(self.Adjacency)):
	    if (self.Adjacency[a,newa]!=0 or self.Adjacency[newa,a]!=0):
	      if connlist[newa]==0:  _Mark(newa)
	return
      _Mark(0)
      return (not 0 in connlist)
     
    def _RandomizeLima(self):	#HELPER: Add a small perturbation to existing links, only necessary when these are directly interpreted as biomass flows to avoid instabilities in the diagonalization routine.
      if (self.P<=0):  return  
      for rg in range(len(self.Adjacency)):
	for li in range(len(self.Adjacency)):
	  if self.Adjacency[rg,li]!=0:  self.Adjacency[rg,li]=np.random.normal(self.Adjacency[rg,li],self.P)
   
##THE HELPER FUNCTIONS TO GENERATE NICHE MODEL TOPOLOGIES! REQUIRES SETTING N,C

    def _SetSpeciesNiche(self): #HELPER: SET NICHE VALUES(n), RANGES(r) and RANGE CENTERS(c), according to connectance
      self.n=np.random.rand(self.N)
      betar=(1.0/(2*self.C)-1.) #X^g=Y, Y is betadist with a=1/g,b=1=>1-Y is beta w. a=1, b=1/g, we require b=1/(2C)-1
      self.r= self.n*(np.ones(self.N) - pow(np.ones(self.N)-np.random.rand(self.N),1./betar));
      self.c= np.asarray([np.random.uniform(self.r[i]/2.,self.n[i]) for i in range(self.N)])
      
      
    def _SetLinks(self):  	#HELPER:Create the adjacency matrix from n,r,c
      
      self.primprod=np.ones(self.N,dtype=np.int8)
      self.toppred=np.ones(self.N,dtype=np.int8)
      links=0
      for i_n,predn in enumerate(self.n):
	sump=0
	for i_prey,preyn in enumerate(self.n):
	  if abs(preyn-self.c[i_n])<self.r[i_n]/2.:  
	    self.Adjacency[i_prey,i_n]=1.
	    self.primprod[i_n]=0; self.toppred[i_prey]=0
	    sump += 1; links+= 1
	  else:
	    self.Adjacency[i_prey,i_n]=0.
	if sump==1 and self.Adjacency[i_n,i_n]!=0:  
	  self.Adjacency[i_n,i_n]=0; self.primprod[i_n]=1; links-=1 #Avoid Cannibalistic Primary Producers
      return links

## CHECKERS IF INPUT PARAMETERS AND FILES ARE SANE   
    def _checkParameters(self):  					##Check if the current paramters are sane (Niche Models)
        if self.N==None or self.C==None:
	  print "N  or C is None!! Are N and C properly initialized?"
	  return False
	if (self.C*self.N)<1:
	  print "WARNING: C is too low to allow a connected Web! Set to C=1/N."
	  self.C=1./self.N
	if (self.C*self.N**2)<self.N*1.25:
	  print "WARNING: C is low, so finding a connected Web can take VERY long."
	print "PARAMETERS ARE:\nSpecies:\t",self.N,"\nConnectance:\t",self.C
	return True


    def _checkAdjacency(self):						##Check if the input Adjacency is sane (Real Webs)
      if not len(self.Adjacency)==len(self.Adjacency[0]):  
	  print "ADJACENCY has to be a square matrix."
	  return False
      if not self._TestConnect():  print "FW Adjacency consists of multiple components."
      return True

##MISC HELPERS FOR THAT MAY BE CALLED BY MORE ADVANCED PARAMETERIZATIONS
    def getTrophicLevels(self):
      flws=self.Adjacency.copy()
      for to in range(len(flws)):
	totflow=sum(flws[:,to])
	if totflow>0:
	  flws[:,to]=flws[:,to]/totflow
	flws[to,to]=-1
      return solve(flws.T,-1.*np.ones(len(flws)))-1.





###########INTERFACE STARTS HERE #################################################################################################
##################################################################################################################################

    
    def makeNicheFW(self,N=None,C=None,P=None):  #MAKE AN ADJACENCY MATRIX! THIS IS THE ONLY PUBLIC FUNCTION THAT SHOULD BE CALLED. IF CALLED WITHOUT PARAMETERS REUSES THE SAME AS LAST CALL
      if N is not None:  self.N=N  #NUMBER OF SPECIES
      if C is not None:  self.C=C  #CONNECTANCE
      if P is not None:  self.P=P  #Perturbation of Topology (if required)
      if self._checkParameters()==False:  return		##THERE IS A PROBLEM AND I WILL NOT MAKE A WEB
            
      self.Adjacency=np.zeros((self.N,self.N))
      done=False	 #It will generate niche webs with given N,C until it finds a connected one (might take a while for large N or small C).
      ct=0
      while (not done):
	self._SetSpeciesNiche()
	links=self._SetLinks()
	if abs(self.C*self.N**2-links)<0.1*self.C*self.N**2:
	  done=self._TestConnect()
	#print self.Adjacency,links,self.C*self.N**2,self._TestConnect()
	ct+=1
	if (ct)%1000==0:  print "So far generated ",ct," unconnected topologies, maybe increase C or decreas N to speed up."
      self._RandomizeLima() #IF NECESSARY RANDOMIZE THE LINKS A BIT (see function descrption)
      print "Found a Connected Niche Web on attempt #",ct
      return self.Adjacency	#RETURNS THE FINAL ADJACENCY MATRIX
    
    
    def loadAdjFile(self,filename,P=None):  #LOAD AN ADJACENCY FROM A FILE
      self.Adjacency=np.loadtxt(filename,comments="#")
      if not self._checkAdjacency(): return
      
      self.N=len(self.Adjacency)  #NUMBER OF SPECIES *NOT NEEDED A PRIORI*
      #self.C=sum([1 for line in Adj for ent in line if ent])  #CONNECTANCE
      if P is not None:  self.P=P  #Perturbation of Topology (if required)
      
      #Provide Primary producer and top predator array
      self.primprod=np.ones(len(self.Adjacency))	
      self.toppred=np.ones(len(self.Adjacency))
      for i in range(len(self.Adjacency)):
	if sum(self.Adjacency[i,:])!=self.Adjacency[i,i]:  self.toppred[i]=0
	if sum(self.Adjacency[:,i])!=self.Adjacency[i,i]:  self.primprod[i]=0      
      
      return self.Adjacency	#RETURNS THE FINAL ADJACENCY MATRIX

##INTERFACE TO GET OR PRINT THE ADJACENCY MATRIX
    def getAdjacency(self):
      return self.Adjacency	#WRAPPER TO RETURN A FINISHED  ADJACENCY IF YOU NEED IT AGAIN
    
    def printAdjacency(self,filename="Adjacency.txt"):
      np.savetxt(filename,self.Adjacency,fmt="%1.5f")	#PRINT THE ADJACENCY TO A FILE WITH NAME <filename> (defaults to Lima.txt)

#############################################################################################################
#############################################################################################################
##HERE ENDS THE CLASS THAT CONTAINS TOPOLOGIES ##############################################################

#############################################################################################################
#HERE STARTS THE CLASS THAT GENERATES A GENERALIZED MODEL FROM A FW TOPOLOGY ################################
class FoodWebModel:
  def __init__(self):
      self.topo=FoodWebAdjacency()						#Empty Topology

##HELPERS####################################################################################################
  def _normalizeFlows(self):						
      for i_cur in range(self.topo.N):						#HELPER: Normalize the feeding parameters. (SHOULD BE CALLED BY ANY FUNCTION SELECTING A MODEL)
	csum=sum(self.chi[i_cur,:])
	bsum=sum(self.beta[:,i_cur])
	if csum>0:  self.chi[i_cur,:]=self.chi[i_cur,:]/csum
	if bsum>0:  self.beta[:,i_cur]=self.beta[:,i_cur]/bsum

  #def makeChiFromFlowEqualization(self,primprodinput=None):
      #if primprodinput==None:  primprodinput=self.topo.primprod.copy()  #If the input flows into the NW are unknown treat them as all equal.
      #extraPP=np.dot(self.beta,primprodinput*sigma)
 
      #inflowmatrix=self.beta.copy()
      #for frm in range(len(inflowmatrix)):
	#inflowmatrix[:,frm]=inflowmatrix[:,frm]*self.sigma[frm]
	
      #inflowmatrix=inflowmatrix-np.diag(np.ones(len(inflowmatrix)))	#inflow_pred=sigma_me*beta_predme*(inflow_me+PP_me)#-extraPP=sigma_me*beta_predme*inflow_me-inflow_pred
      #flows=solve(inflowmatrix,-extraPP)	#The inflows of each species
      
      #for pred in range(len(self.beta)):
        #for prey in range(len(self.beta)):
	    #self.chi[pred,prey]=self.sigma[prey]*beta[pred,prey]*(flows[prey]+primprodinput[prey])/flows[pred]   

  
  def makeBetaFromFlowEqualization(self,M=None):				
      if M==None:  M=self.topo.toppred.copy().T  #If the output flows of sinks are unknown treat them as all equal.
      Mchi=np.dot(M,self.chi)
						#OUT_i=IN_i=>L_k=sum_i (L_i-PP_i+M_i) chi_ik #For TP i: L_i=0 #FOR PP i: chi_i=0
						#=>L_k=sum_i (L_i+M_i) chi_ik=sum_i 1/sigma_i L_i chi_ik + sum_tp M_tp chi_tp,k
						#=>-dot(M,chi)=-diag(L)+(L_i,chi_ik/sigma_i)
      sigmachi=self.chi.copy()		#Measures the outflow relative to the outflow of each of my predators
      for pred in range(len(sigmachi)):
	  if self.sigma[pred]>0:  sigmachi[pred,:]=sigmachi[pred,:]/self.sigma[pred]
	    #
      sigmachi=sigmachi-np.diag(np.ones(len(sigmachi)))	#Primprod+outflow_prey=inflow, outflow_prey=inflow_prey*sigma_prey
      flows=solve(sigmachi.T,-Mchi)				#The outflows of each species due to Predation
    #  print flows
    #  print self.sigma
    #  print M
      for pred in range(len(self.beta)):
        for prey in range(len(self.beta)):
	  InF=M[pred]
	  if self.sigma[pred]>0:  InF+=flows[pred]/self.sigma[pred]
	  if self.chi[pred,prey]>0: self.beta[pred,prey]=InF*self.chi[pred,prey]/flows[prey]

      
## MAKE THE JACOBIAN AND MEASURE STABILITY
  def makeJacobian(self):				#COPIED FROM C PROGRAM: CALCULATES AND RETURNS THE JACOBIAN MATRIX FOR A PARAMETRIZED MODEL
      self.Jacobian=np.zeros((self.topo.N,self.topo.N))
      N=self.topo.N
		   
      for n in range(N):
	for i in range(N):
	  if (i != n): 		# off diagonal
	    dsum = 0; # for calculating the sum in the loss by predation part for mutualistic effects
	    for m in range (N):
		dsum +=  self.beta[m][n] * self.lambdas[m][i] * (self.gamma[m] - 1.) * self.chi[m][i]
	    self.Jacobian[n][i] = self.alpha[n]*(self.rho[n] * self.gamma[n] * self.chi[n][i] * self.lambdas[n][i] - self.sigma[n] * (self.beta[i][n]*self.psi[i] + dsum) )# gain by predation -  loss by predation
  
      for i in range(N):	# diagonales
        dsum = 0;
        for m in range(N):
	    dsum  +=self.beta[m][i] * self.lambdas[m][i] * ( (self.gamma[m] - 1.) * self.chi[m][i]  +1. );
        
	self.Jacobian[i][i] = self.alpha[i]*( (1.-self.rho[i])*self.phi[i] + self.rho[i] * (self.gamma[i] * self.chi[i][i]  * self.lambdas[i][i] + self.psi[i]) - (1.-self.sigma[i]) * self.mu[i]- self.sigma[i] * ( self.beta[i][i]*self.psi[i] + dsum ) )                         # primary production + gain by predation  - mortality- loss by predation
	
      return self.Jacobian




						 
#####INTERFACE################################################################################################
##############################################################################################################

## FOR CONVENIENCE: GET AN ADJACENCY INTO self.topo     
  def makeNicheFW(self,N,C,P=None):						#Make a Niche Model Topology 
    self.topo.makeNicheFW(N,C,P)

  def loadAdjFile(self,fname,P=None):						#Load A Real Topology 
    self.topo.loadAdjFile(fname,P)

##GET INFORMATIONS ABOUT THE STABILITY OF THE CHOSEN PARAMETRIZATION


  def testStability(self):				#TESTS IF THE CHOSEN MODEL ALLOWS FOR A STABLE STEADY STATE. Returns True or False 
    if max(eigvals(self.Jacobian))<-1e-6:  return True
    return False

  def getEigenvalues(self):				#TESTS IF THE CHOSEN MODEL ALLOWS FOR A STABLE STEADY STATE. Returns True or False 
    return sorted(eigvals(self.Jacobian))

## PRINTING TO FILES  
  def printParameters(self,filename="Params.txt"):	#CONVENIENT PRINTER FOR THE CHOSEN PARAMETERS. Prints them to <filename>, default "Params.txt"
	  f=open(filename,'w')
	  if self.topo.c!=None:    #Is it a niche model topology
	    f.write("#N\talpha\tgamma\tmu\tphi\tpsi\trho\tsigm\tni\tci\tri\n")
	    for i in range(self.topo.N):
	      f.write("%d\t%1.3f\t%1.3f\t%1.3f\t%1.3f\t%1.3f\t%1.3f\t%1.3f\t%1.3f\t%1.3f\t%1.3f\n" % (i, self.alpha[i],self.gamma[i],self.mu[i],self.phi[i],self.psi[i],self.rho[i],self.sigma[i],self.topo.n[i],self.topo.c[i],self.topo.r[i]))
	  else:  
	    f.write("#N\talpha\tgamma\tmu\tphi\tpsi\trho\tsigm\n")
	    for i in range(self.topo.N):
	      f.write("%d\t%1.3f\t%1.3f\t%1.3f\t%1.3f\t%1.3f\t%1.3f\t%1.3f\n" % (i, self.alpha[i],self.gamma[i],self.mu[i],self.phi[i],self.psi[i],self.rho[i],self.sigma[i]))
	      
	      
	  f.write("#From\tTo\tchi[to]\tbeta[fr]\tlambda\n")
	  for frm in range(self.topo.N):
	    for to in range(self.topo.N):
	      if self.chi[to,frm]!=0:
		f.write("%d\t%d\t%1.3f\t%1.3f\t%1.3f\n" % (frm,to,self.chi[to,frm],self.beta[to,frm],self.lambdas[to,frm]))
	  f.close()
      
  def printJacobian(self,filename="Jacobian.txt",delimiter="\t"):    	#CONVENIENT PRINTER FOR THE JACOBIAN MATRIX. Prints it to <filename>, default "Jacobian.txt"
      np.savetxt(filename,self.Jacobian,fmt="%1.5f",delimiter="\t")

  def printAdjacency(self,filename="Lima.txt",delimiter="\t"):		#Wrapper to print the Topology
    self.topo.printAdjacency(filename)


##DIFFERENT PARAMETRIZATIONS: CHANGE THIS TO YOUR OWN DESIRED PARAMETRIZATION BY SETTING THE PARAMETERS
##############################################################################################################
  def makeDummyModel(self,morti=1.):				#SETTING UP A DUMMY MODEL!!
      self.alpha=np.ones(self.topo.N)			# vector: time scales
      self.gamma=0.95*np.ones(self.topo.N)		# vector: sensitivity of predation rate to prey density
      self.lambdas=np.ones((self.topo.N,self.topo.N))	# matrix: contribution between prey (one for passive swiching)
      self.mu=np.ones(self.topo.N)			# vector: sensitivity of nonpredation mortality
      self.phi=0.5*np.ones(self.topo.N)			# vector: sensitivity of primary production (nutrient availability)
      self.psi=np.ones(self.topo.N)			# vector: sensitivity of predation rate to predator density
     
      self.rho=self.rho=1.-self.topo.primprod.copy()  	# vector: fraction of biomass gained by predation # 0 predation gain for primprod, 1 for others
      self.sigma=np.asarray([(sum(self.topo.Adjacency[i_s,:]))/(morti+sum(self.topo.Adjacency[i_s,:])) for i_s in range(len(self.topo.Adjacency[0]))])
							# vector: fraction of biomass loss by modelled predation -> set by biomass flows in Lima + mortality loss
      self.beta=self.topo.Adjacency.T.copy()			# matrix: beta (loss rate of j by i predation) -> Set by biomass flows in Lima
      self.chi=self.topo.Adjacency.T.copy()			# matrix: chi (feeding rate of j on i predation) -> Set by biomass flows in Lima
      self._normalizeFlows()				#Now normalize them
  
  def makeRandomModel(self):     			#SETTING UP A MODEL WITH RANDOM PARAMETERS!!
      self.alpha=np.random.uniform(0.01,1.,size=self.topo.N)		#Alpha Uniform range
      self.gamma=np.random.uniform(0.5,1.5,size=self.topo.N)		#Gamma Uniform range
      self.lambdas=np.ones((self.topo.N,self.topo.N))			#Passive Prey switiching
      self.mu=np.random.uniform(1.,2.,size=self.topo.N)			#EUniform mortality exponent
      self.phi=np.random.uniform(0.,1.,size=self.topo.N)		#Uniform prim.prod exponent
      self.psi=np.random.uniform(0.5,1.5,size=self.topo.N)		# Uniform predation exponent
      self.rho=1.-self.topo.primprod.copy()				# 0 predation gain for primprod, 1 for others
      self.sigma=(1.-self.topo.toppred.copy())*np.random.uniform(0.25,1.,size=self.topo.N) #predation loss is 0 for top predators, uniform else
      
      self.chi=np.zeros((self.topo.N,self.topo.N))
      self.beta=np.zeros((self.topo.N,self.topo.N))
      for i_pred in range(self.topo.N):					#Assign values to the nonzero feeding parameters. Can also be done by directly using biomass flows.
	for i_prey in range(self.topo.N):
	  if self.topo.Adjacency[i_prey,i_pred]!=0:
	    self.chi[i_pred,i_prey]=np.random.uniform(0.1,1.0)		#Note the inverse order of chi,beta and lima!!
            self.beta[i_pred,i_prey]=np.random.uniform(0.1,1.0)
      self._normalizeFlows()
      
  ##FUNCTIONS TO MODIFY THE DUMMY AND RANDOM MODEL ABOVE
  def makeRateScalingbyTL(self,alpharat=0.08):		      #Scale Rates with TrophicLevels
      self.alpha=alpharat**(self.topo.getTrophicLevels()/4.)  #Scaling of alpha with BodyMass alpha~M**(-1/4), M~12.5^TL=>alpha=0.08**(TL/4)
  
  def makeRateScalingbyBM(self,bodymass,prefactor=0.1):		#Scale the turnover rates based on body mass.
      if not isinstance(bodymass,(np.ndarray,list)):  print "ERROR: bodymass is not a vector."	#Scale Rates with BodyMasses
      elif len(bodymass)!=len(self.alpha):  print "ERROR: bodymass Vector must fit system dimensions"	#Scale Rates with BodyMasses
      else:  self.alpha=(prefactor*np.asarray(bodymass))**(-0.25)  #Scaling of alpha with BodyMass alpha~M**(-1/4), M~12.5^TL=>alpha=0.08**(TL/4)
  
  def makeRateScalingbyNiche(self,alpharat=0.08):
      self.alpha=alpharat**self.topo.n					#Allometric scaling with niche values 
 
  def makeBetaFromChi(self,netoutput=None):				      #Use conservation of mass to set betas (requires that chi has been set!)
    self.makeBetaFromFlowEqualization(netoutput)				      #Netoutput can be an array containing net output flows of the top predators if available, else they are set to 1.

  def adjustModelElasticities(self,gamma=None,lambdas=None,mu=None,phi=None,psi=None):				#ADJUST ELATICITIES MANUALLY:
      def setParameter(newval,toset):								#How to handle different inputs:
	if newval==None:  return toset								#None=Not given: set to the old value
	if isinstance(newval,(int, long, float)):  return newval*np.ones_like(toset)		#Number: set all equal to this number
	if isinstance(newval,list):  newval=np.asarray(newval)					#list->convert to array
	if isinstance(newval,np.ndarray):							#array:check if shape is consistend, then set to the input array
	  if newval.shape==toset.shape:  return newval
	  else:  print "Parameter Requires an array of shape "+str(toset.shape)+" but obtained shape"+str(newval.shape)
	print "Could not find a matching scenario to set elasticity, check your input."
	return toset										#if not one of these types simply retain the old value
      
      self.gamma=setParameter(gamma,self.gamma)		
      self.lambdas=setParameter(lambdas,self.lambdas)	# matrix: contribution between prey (one for passive swiching)
      self.mu=setParameter(mu,self.mu)			# vector: sensitivity of nonpredation mortality
      self.phi=setParameter(phi,self.phi)		# vector: sensitivity of primary production (nutrient availability)
      self.psi=setParameter(psi,self.psi)		# vector: sensitivity of predation rate to predator density
     
   


  
  